'''
constants.py - physics constants
'''

from euclid import Point3, Vector3

ORIGIN = Point3(0.,0.,0.)

# control responsiveness

TURN_SPEED = 180. # degrees / second

ACCELERATION = 9.0 # units/second^2, I guess


# friction values are velocity multipliers per second (1.0 means no friction, 0.0 means freeze to a stop)
# (0.01 is not as extreme a value as it looks, due to the exponential when it's used)

TERRAIN_FRICTION = 0.85

BRAKING_FRICTION = 0.01

TERRAIN_IMPACT_DAMPING = 0.000001 # same as "0.1 over 1/6 second".
    # Note: effect of a given value of TERRAIN_IMPACT_DAMPING depends on mushiness of potential,
    # so when that's modified, this value also probably needs to be modified.


# jumping and gravity

# set these...
JUMP_HEIGHT = 3.
JUMP_DURATION = 1.

# to compute these:
# (using d = 0.5 A t^2, i.e. JUMP_HEIGHT == 0.5 GRAVITY (JUMP_DURATION/2.)**2,
#  and v = a dt, i.e. JUMP_SPEED == GRAVITY (JUMP_DURATION/2.))
GRAVITY = JUMP_HEIGHT / (0.5 * (JUMP_DURATION/2.)**2)
JUMP_SPEED = GRAVITY * (JUMP_DURATION/2.)
# print "GRAVITY = %r, JUMP_SPEED = %r" % (GRAVITY, JUMP_SPEED)

UP = Vector3(0.0, 1.0, 0.0) # see also player._up
JUMP_VELOCITY = JUMP_SPEED * UP # XXX later, will depend on local terrain etc, see comment below


# terrain potential function

BALL_RADIUS = 1. # WARNING: this is redundant with a value which is passed in
    # (to our potential-computing method) which it will ignore (for speed),
    # which comes from a constant in Player; and is also hardcoded into the Sphere shape.

MESH_DENSITY = 2 # the value of K passed to certain methods

# Let's push hard enough to balance gravity with a single meshpoint inside
# the ball at a depth of half-radius. At that depth there'd actually be
# more like 10 points inside it (with K == 2), so the equilibrium depth
# should be much less, even without knowing the functional form.
# If we increase K**2 we should reduce the force per meshpoint,
# but we'll never increase it very much due to speed, so nevermind.

# And as for functional form, we want it fast to compute,
# so try:
#
#    magnitude_divided_by_length = (radius ** 2 - distance ** 2) ** HARDNESS_EXPONENT * BALL_FORCE_COEFFICIENT
#
#    magnitude = magnitude_divided_by_length * length # implicit in not normalizing vec
#
# and then:
#
#    force = vec * magnitude_divided_by_length    # avoid normalize, for speed
#
#    or equivalently
#
#    force = (vec / length) * magnitude      # clearer, since vec is normalized
#
# which at half-radius distance should give us GRAVITY, re above discussion.
# It should also work ok (no error and smooth) even at vec == 0,0,0 or edge of ball (giving 0 vec either way).

HARDNESS_EXPONENT = 2 # try this first -- also try values from 1 through 5 if necessary
    # (some of this, maybe (guess) 0.5 or so, is to compensate for the _divided_by_length kluge)

# solve for this (using private temp constants):
_test_depth = BALL_RADIUS / 2.0
_test_length = BALL_RADIUS - _test_depth
# we want GRAVITY == abs(vec) * magnitude_divided_by_length, so:
_mag_over_length = GRAVITY / _test_length
# now solve for BALL_FORCE_COEFFICIENT from _mag_over_length
BALL_FORCE_COEFFICIENT = _mag_over_length / ( (BALL_RADIUS ** 2 - _test_length ** 2) ** HARDNESS_EXPONENT )

print "BALL_FORCE_COEFFICIENT =", BALL_FORCE_COEFFICIENT ####

