
from pyglet.gl import *
import pyglet
from pyglet.window import key

from math import pi, sin, cos, atan2

from shapes import Sphere

from euclid import Vector3, Point3, Quaternion

from constants import *
    # this includes:
    ## ORIGIN = Point3(0.0, 0.0, 0.0) ### should be aspect of world or terrain; has multiple defs, some with int vs float XXX

# ==

INIT_PLAYER_POS = ORIGIN + Vector3(2.0, 1.0, 2.0) # XXX or lower (let collision raise it to floor)
INIT_PLAYER_HEADING = Vector3(1.0, 0.0, 1.0).normalized()

def heading_to_angle(x,y):
    return atan2(y, x)

# ==

class MovingGroup(pyglet.graphics.Group):
    "draw our vertices translated and rotated by a mutable coordinate system"
    # XXX do we need __eq__? not if we just cache the ones with same parent and mover.
    def __init__(self, parent, mover):
        "uses mover.pos and mover.quat"
        pyglet.graphics.Group.__init__(self, parent) #k assume init arg is parent group
        self.mover = mover # knows position, which we use in set_state
    def set_state(self):
        glPushMatrix() # assume GL_MODELVIEW
        p = self.mover.pos
        glTranslatef(p.x, p.y, p.z)
        q = self.mover.quat
        angle, axis = q.get_angle_axis()
        glRotatef(angle * 180. / pi, *axis )
    def unset_state(self):
        glPopMatrix()

# ==

class Player:
    
    _up = UP
    keyboard = None # should be set in .update (kluge) to a KeyStateHandler
    debug = True ###
    direction_vlist = None
    ENABLE_COLLISIONS = 3 # can be toggled externally for debug; chooses one of several collision models
    
    def __init__(self, initpos = INIT_PLAYER_POS, other = False):
        self.other = other # true means not the main character; affects controls
        self.radius = 1.0 # XXX pass to Sphere
        self.shape = Sphere()
        self.pos = initpos.copy() # XXX store this in a collision body? nah, we are one
        self.lastpos = self.pos.copy() # for verlet integration -- ASSUMES ZERO INITIAL VELOCITY AND ACCELERATION
        self.lastdt = 1. # for time-corrected verlet integration (initial value can be arbitrary nonzero, since initial V & A is 0)
            # (of course the initial A from the terrain is *not* really 0 -- we'll have to see whether this matters;
            #  if it does, see time_corrected_verlet_newpos for what to do about it)
        self.quat = Quaternion() # rotation (owned and mutable)
        self.heading = INIT_PLAYER_HEADING.copy()
        self.angle = heading_to_angle(self.heading.x, self.heading.z) # angle of heading (xz plane), radians
        self.velocity = Vector3(0., 0., 0.) # only used when not ENABLE_COLLISIONS (I hope)
        self.ground_speed = 0.
        self.roll_rate = 0.
        self.roll_axis = Vector3(0., 0., 0.)
        self.jump = False
        self._do_add_player = False
        print "terrain-collisions:", self.collision_model_short_phrase()
        return

    def control_add_player(self):
        self._do_add_player = True
    
    def time_corrected_verlet_newpos(self, dt, A, simple_friction_dt = 1.0):
        """
        Given dt and total acceleration (from terrain, controls, friction, gravity),
        and an optional velocity multiplier for simple friction,
        return desired new self.pos, without setting it in self.
        (Later, caller should pass it to time_corrected_verlet_update, to set it.)
        """
        # http://www.gamedev.net/reference/programming/features/verlet/
        # A Simple Time-Corrected Verlet Integration Method
        # by Jonathan Dummer 
        #
        # Time-Corrected Verlet:
        # xi+1 = xi + (xi - xi-1) * (dti / dti-1) + a * dti * dti
        # 
        # starting condition derivation (if initial V or A was nonzero):
        # xi - xi-1 = vi-1 * dti-1 + 0.5 * ai-1 * dti-1 * dti-1
        #
        # Note: consequences of letting controls vary this accel quickly are not known
        # (the derivation of verlet integration assumes it won't change quickly),
        # but are probably ok.
        pos = self.pos
        newpos = ( pos
                   + (pos - self.lastpos) * (dt / self.lastdt) * simple_friction_dt
                   + A * dt * dt
                  )
        return newpos

    def time_corrected_verlet_update(self, dt, newpos):
        self.lastpos = self.pos
        self.pos = newpos
        self.lastdt = dt
        return

    def terrain_accel(self, pos, terrain, debug = False): # XXX make this realistic re terrain, walls, etc # XXX other balls??
        "return the acceleration vector on self (at pos) due to collision force from the terrain"
        # note: caller will also be able to use this to understand friction, if it makes some assumptions
        # note: assume our mass is 1, so we can just compute and return force, not that it matters
        # (since all numeric constants are pretty arbitrary)

        assert self.ENABLE_COLLISIONS

        if self.ENABLE_COLLISIONS == 3:
            # the real potential function, we hope
            return terrain.force_on_ball( pos, self.radius, debug = debug)

        if self.ENABLE_COLLISIONS == 1:
            # initial stub for testing: upward if too low (lower means stronger), and that's all.
            ground_height = 2.5
            equilibrium_dist_under_limit = 0.5
        elif self.ENABLE_COLLISIONS == 2:
            # slightly more realistic stub -- actually notice the terrain height.
            # (not very realistic on slopes! 1. force never points sideways 2. ballsides intersect hillsides)
            ground_height = terrain.terrain_height(pos)
            equilibrium_dist_under_limit = 0.25
        if pos.y < ground_height:
            force = (ground_height - pos.y) * UP * (GRAVITY/equilibrium_dist_under_limit)
            # print "force", force ###
        else:
            force = Vector3(0.0, 0.0, 0.0)
        return force
    
    def camera_pos(self, zoom):
        zoom = 1.0 / zoom
        return self.pos + (self._up * 2.0 - self.heading * 6.0) * zoom

    def lookat(self, zoom):
        up = self._up
        center = self.pos
        eye = self.camera_pos(zoom)
        gluLookAt(eye.x, eye.y, eye.z,
                  center.x, center.y, center.z,
                  up.x, up.y, up.z, )

    def control_jump(self):
        "called externally on each jump key press"
        if self.ENABLE_COLLISIONS: # otherwise, no way to get down, though braking stops you high up
            self.jump = True

    def debug_toggle_collisions(self):
        self.ENABLE_COLLISIONS = (1 + self.ENABLE_COLLISIONS) % 4 # 1 false value, 3 true values (2nd/3rd ones are most interesting)
        print "terrain-collisions:", self.collision_model_short_phrase()
        self.velocity = Vector3(0,0,0)
        self.lastpos = self.pos.copy()

    def collision_model_short_phrase(self):
        return [  "disabled",
                  "imaginary plane",
                  "real terrain but fake non-conservative potential",
                  "real terrain and potential"] [self.ENABLE_COLLISIONS]

    def compute_control_accel(self, dt, free_fall, terrain_force, fast):
        """
        Return the acceleration vector due to controls (plus non-simple friction).
        (Leave out simple friction (proportional to velocity) and gravity.)
        """
        kb = self.keyboard

        if not free_fall:
            acceleration = 0.0
            if kb[key.UP]:
                acceleration += ACCELERATION
            if kb[key.DOWN]:
                acceleration -= ACCELERATION
            if fast:
                acceleration *= 2.

            # XXX make it higher if velocity is low, i.e. a strong pushoff from standing start?
            # XXX if ground is not flat, what do we do -- tilt the plane in which to determine this?
            control_accel = (self.heading * acceleration)
                ### XXX also needs "rotated tire effect" for better steering (accel transverse to velocity and "up");
                # friction; ... [see below for more comments on this]

            # XXX jump should depend on ground normal and maybe on "steering"
            if self.jump:
                control_accel += JUMP_VELOCITY / dt

            # heavy damping when we hit the ground (down or up V component, when touching it)
            #### XXX TODO, bugfix: only on down, not up -- otherwise it makes jump too low
            ground_normal = terrain_force.normalized() # or could pass it in
            est_last_velocity = (self.pos - self.lastpos) / self.lastdt # dup code with time_corrected_verlet_newpos
            impact_V = min(0., est_last_velocity.dot(ground_normal)) * ground_normal # component perp to ground (down direction only)
            damp_this_part_out = impact_V * (1. - TERRAIN_IMPACT_DAMPING ** dt)
            control_accel -= damp_this_part_out / dt
            pass
        else:
            control_accel = Vector3(0.,0.,0.)
        
        self.jump = False # whether or not free_fall

        ### XXX TODO: non-simple friction (e.g. complex braking, steering == "rotated tire effect").
            # XXX if speed small, set to zero and also "hold still"? (parking brake on slope, and full stopping)

        return control_accel

    def compute_simple_friction(self, free_fall, braking, dt):
        "return the velocity multiplier per second (1.0 for no friction, 0.0 for complete frozen stop)"
        # XXX depend on ground_speed? BUG: ok for ground speed, not up/down damping
        if free_fall:
            return 1.0
        else:
            if braking:
                return BRAKING_FRICTION ** dt
            else:
                return TERRAIN_FRICTION ** dt

    paused = False
    
    def pause_and_print_player_pos(self):
        self.paused = not self.paused
        if self.paused:
            print "pausing with player at %r" % self.pos
        else:
            print "resuming"
    
    def update(self, dt, terrain, kb):
        assert dt
        # note: kb can be false even if it's been initialized! It's a dict-like object. Once it's been used,
        # it seems not to be false ever again (guess, not confirmed -- maybe it has explicit keys with False values then?).
        if kb is None:
            print "update with no keyboard -- should never happen"
            return
        
        self.keyboard = kb # kluge, for access by our submethods
        
        ## fast = kb[key.F] # F = fast (not very useful -- hard to remember, nonintuitive to use)
        # note: key.F is now for toggling fullscreen mode. [old meaning of fast is now hardcoded always on in constants.]
        fast = False
        
        turn = 0.0
        if kb[key.LEFT]:
            turn += TURN_SPEED
        if kb[key.RIGHT]:
            turn -= TURN_SPEED
        if fast:
            turn *= 1.5
        turn *= dt # number of degrees to rotate heading by

        # change heading (controls camera, and direction of control-acceleration)
        self.angle += turn * pi / 180.
        self.heading = Vector3(sin(self.angle), 0.0, cos(self.angle)) # assert UP is +y

        if not self.paused:
            # pause helps with screenshots; we let turning (above) still work
            # when paused, as well as lawn-modification (after this 'if')
        
            braking = kb[key.B] # used below (when not free_fall)
                # XXX braking sound (when actually used)

            # terrain collision, gravity, position integration
            # part 1: compute A_ter and free_fall
            if self.ENABLE_COLLISIONS: # can be toggled as debug setting (key.E)
                self.velocity = None # make sure this obsolete attr is not used in this case
                pos = self.pos
                A_ter = self.terrain_accel(pos, terrain) # accel due to terrain contact force
                free_fall = (abs(A_ter) == 0.) # future: might differ if we bounce off "wall" -- not sure
                    # if false, we're touching terrain (in general, intersecting it)
            else:
                # old code (part 1), for debugging
                A_ter = self._up
                free_fall = False

            # part 2: compute things from A_ter
            ground_normal = A_ter.normalized() # used below
            terrain_force = A_ter # assume self's mass is 1
            control_accel = self.compute_control_accel(dt, free_fall, terrain_force, fast) # includes non-simple friction
            simple_friction_dt = self.compute_simple_friction(free_fall, braking, dt)
            
            # part 3: gravity, position integration (with simple_friction)
            if self.ENABLE_COLLISIONS:
                control_accel -= self._up * GRAVITY
                A = A_ter + control_accel # total accel
                newpos = self.time_corrected_verlet_newpos(dt, A, simple_friction_dt)
                displacement = newpos - self.pos # used below
                ## if abs(displacement) > 0.5:                
                ##     print "moved", abs(displacement) # rare but happens, but very rare to be more than 1.0
                self.time_corrected_verlet_update(dt, newpos) # sets self.pos and some other variables
            else:
                # no gravity in this case (since no ground to push up when it gets too low)
                self.velocity += control_accel * dt
                self.velocity *= simple_friction_dt
                displacement = self.velocity * dt
                self.pos += displacement

            # ball-rolling appearance        
            if not free_fall:
                # change axis and rate of ball-rolling to match motion along ground
                # (visually, I often think it's rolling too fast, but AFAICT it's accurate,
                #  e.g. when I compare ball surface features with lawn wireframe features.
                #  Would a shadow help?? Actually I just fixed a bug in rolling, and maybe
                #  this has fixed that impression... so maybe i was right.)
                
                ground_displacement = displacement - displacement.dot(ground_normal) * ground_normal
                direction_along_ground = ground_displacement.normalized()
                self.roll_axis = ground_normal.cross(direction_along_ground)
                ground_distance = abs(ground_displacement)
                roll_angle = ground_distance / self.radius # in radians -- this value makes it appear to roll on the surface
                self.roll_rate = roll_angle / dt
            else:
                # roll axis is unchanged, roll rate gradually reduced as if by air friction (for testing, use .3 for not so gradually)
                # XXX might be better to change axis/rate to indicate desired acceleration,
                # as if preparing to do that once touching the ground; not yet, since its control on ground doesn't work that way
                self.roll_rate *= (.9)**dt
                ground_distance = 0.0 # needed below
            
            roll_angle = self.roll_rate * dt
            if abs(self.roll_axis):
                self.rotate(roll_angle, self.roll_axis)

            # speed indicator
            if 0:
                ground_speed = ground_distance / dt
                if int(ground_speed*100) != int(self.ground_speed*100):
                    print ground_speed, "\r",
                    import sys
                    sys.stdout.flush()
                self.ground_speed = ground_speed

            # debugging indicators
            if self.direction_vlist:
                pos = self.pos
                ## end = self.pos + axis * 3.0 # axis end
                end = self.pos - self._up * 10.0
                self.direction_vlist.vertices[:] = (pos.x, pos.y, pos.z, end.x, end.y, end.z)

        # tell caller how to affect world, by return value # XXX depend on free_fall?
        what = dict(
            make_hill = kb[key.H],
            make_depression = kb[key.D], # "dig"
            recolor = kb[key.C] ## XXX maybe: include the color to use
         )
        if self._do_add_player:
            self._do_add_player = False
            where = self.pos + self.heading * 5
            what['add_player'] = where ## XXX maybe: (where, (0,0,200,255))
        return what

    def get_help_text(self): # XXX CALL ME
        # commented out stuff in the string is NIM; whole thing may not be up to date, eg lacks F8
        return """
            acceleration - arrow keys # or AWSD
            brake - B
            make hill - H
            make depression - D # or M for "mow"??  no, that'll be "mole"...
            color - C
            # jump - space
            # look around - mouse motion
            zoom - mouse wheel # , or 1 or 2
        """
    
    def rotate(self, angle, axis):
        "rotate angle (in radians) around axis"
        ## self.quat.rotate_axis(angle, axis)
        ## self.quat = self.quat * Quaternion.new_rotate_axis(angle, axis) # should be and is equivalent, still wrong
        self.quat = Quaternion.new_rotate_axis(angle, axis) * self.quat
            # in theory this is backwards, but it makes rolling correct! bug in euclid, or in my understanding?
        
    def add_to_batch(self, batch, group):
        mygroup = MovingGroup(group, self) # self is the mover, has .pos and .quat
        self.shape.add_to_batch(batch, mygroup)
        self.shape.vertex_list # could be used to move the player, but using the Group is easier

        if self.debug:
            self.direction_vlist = batch.add(2, GL_LINES, None, 'v3f')
    pass
